# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['doorstop',
 'doorstop.cli',
 'doorstop.core',
 'doorstop.core.validators',
 'doorstop.core.vcs',
 'doorstop.gui',
 'doorstop.server']

package_data = \
{'': ['*'],
 'doorstop': ['views/*'],
 'doorstop.core': ['files/*',
                   'files/templates/html/doorstop/*',
                   'files/templates/latex/*']}

install_requires = \
['bottle>=0.12.13,<0.13.0',
 'markdown>=3.3.6,<4.0.0',
 'openpyxl>=3.0.7',
 'plantuml-markdown>=3.4.2,<4.0.0',
 'python-frontmatter>=1.0,<2.0',
 'python-markdown-math>=0.6,<0.7',
 'pyyaml>=5.1,<6.0',
 'requests>=2.0,<3.0',
 'six']

entry_points = \
{'console_scripts': ['doorstop = doorstop.cli.main:main',
                     'doorstop-gui = doorstop.gui.main:main',
                     'doorstop-server = doorstop.server.main:main']}

setup_kwargs = {
    'name': 'doorstop',
    'version': '3.0b7',
    'description': 'Requirements management using version control.',
    'long_description': '[![Linux Test](https://github.com/doorstop-dev/doorstop/actions/workflows/test-linux.yml/badge.svg)](https://github.com/doorstop-dev/doorstop/actions/workflows/test-linux.yml)\n[![Mac OS X Test](https://github.com/doorstop-dev/doorstop/actions/workflows/test-osx.yml/badge.svg)](https://github.com/doorstop-dev/doorstop/actions/workflows/test-osx.yml)\n[![Windows Test](https://github.com/doorstop-dev/doorstop/actions/workflows/test-windows.yml/badge.svg)](https://github.com/doorstop-dev/doorstop/actions/workflows/test-windows.yml)\n<br>\n[![Coverage Status](https://img.shields.io/codecov/c/gh/doorstop-dev/doorstop)](https://codecov.io/gh/doorstop-dev/doorstop)\n[![Scrutinizer Code Quality](http://img.shields.io/scrutinizer/g/doorstop-dev/doorstop.svg)](https://scrutinizer-ci.com/g/doorstop-dev/doorstop/?branch=develop)\n[![PyPI Version](http://img.shields.io/pypi/v/Doorstop.svg)](https://pypi.org/project/Doorstop)\n<br>\n[![Gitter](https://badges.gitter.im/doorstop-dev/community.svg)](https://gitter.im/doorstop-dev/community)\n[![Google](https://img.shields.io/badge/forum-on_google-387eef)](https://groups.google.com/forum/#!forum/doorstop-dev)\n[![Best Practices](https://bestpractices.coreinfrastructure.org/projects/754/badge)](https://bestpractices.coreinfrastructure.org/projects/754)\n\n# Overview\n\nDoorstop is a [requirements management](http://alternativeto.net/software/doorstop/) tool that facilitates the storage of textual requirements alongside source code in version control.\n\n<img align="left" width="100" src="https://raw.githubusercontent.com/doorstop-dev/doorstop/develop/docs/images/logo-black-white.png"/>\n\nWhen a project leverages this tool, each linkable item (requirement, test case, etc.) is stored as a YAML file in a designated directory. The items in each directory form a document. The relationship between documents forms a tree hierarchy. Doorstop provides mechanisms for modifying this tree, validating item traceability, and publishing documents in several formats.\n\nDoorstop is under active development and we welcome contributions.\nThe project is licensed as [LGPLv3](https://github.com/doorstop-dev/doorstop/blob/develop/LICENSE.md).\nTo report a problem or a security vulnerability please [raise an issue](https://github.com/doorstop-dev/doorstop/issues).\nAdditional references:\n\n- publication: [JSEA Paper](http://www.scirp.org/journal/PaperInformation.aspx?PaperID=44268#.UzYtfWRdXEZ)\n- talks: [GRDevDay](https://speakerdeck.com/jacebrowning/doorstop-requirements-management-using-python-and-version-control), [BarCamp](https://speakerdeck.com/jacebrowning/strip-searched-a-rough-introduction-to-requirements-management)\n- sample: [Generated HTML](http://doorstop-dev.github.io/doorstop/)\n\n\n# Setup\n\n## Requirements\n\n* Python 3.7+\n* A version control system for requirements storage\n\n## Installation\n\nInstall Doorstop with pip:\n\n```sh\n$ pip install doorstop\n```\n\nor add it to your [Poetry](https://poetry.eustace.io/) project:\n\n```sh\n$ poetry add doorstop\n```\n\nAfter installation, Doorstop is available on the command-line:\n\n```sh\n$ doorstop --help\n```\n\nAnd the package is available under the name \'doorstop\':\n\n```sh\n$ python\n>>> import doorstop\n>>> doorstop.__version__\n```\n\n# Usage\n\nSwitch to an existing version control working directory, or create one:\n\n```sh\n$ git init .\n```\n\n## Create documents\n\nCreate a new parent requirements document:\n\n```sh\n$ doorstop create SRD ./reqs/srd\n```\n\nAdd a few items to that document:\n\n```sh\n$ doorstop add SRD\n$ doorstop add SRD\n$ doorstop add SRD\n```\n\n## Link items\n\nCreate a child document to link to the parent:\n\n```sh\n$ doorstop create HLTC ./tests/hl --parent SRD\n$ doorstop add HLTC\n```\n\nLink items between documents:\n\n```sh\n$ doorstop link HLTC001 SRD002\n```\n\n## Publish reports\n\nRun integrity checks on the document tree:\n\n```sh\n$ doorstop\n```\n\nPublish the documents as HTML:\n\n```sh\n$ doorstop publish all ./public\n```\n',
    'author': 'Jace Browning',
    'author_email': 'jacebrowning@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/Doorstop',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# SPDX-License-Identifier: LGPL-3.0-only

"""Finding external references."""

import linecache
import os
import re

from doorstop import common, settings
from doorstop.common import DoorstopError

log = common.logger(__name__)


class ReferenceFinder:
    """Finds files referenced from an Item."""

    @staticmethod
    def find_ref(ref, tree, item_path):
        """Get the external file reference and line number.

        :raises: :class:`~doorstop.common.DoorstopError` when no
            reference is found

        :return: relative path to file or None (when no reference
            set),
            line number (when found in file) or None (when found as
            filename) or None (when no reference set)

        """

        # Search for the external reference
        log.debug("searching for ref '{}'...".format(ref))
        pattern = r"(\b|\W){}(\b|\W)".format(re.escape(ref))
        log.trace("regex: {}".format(pattern))  # type: ignore
        regex = re.compile(pattern)
        for path, filename, relpath in tree.vcs.paths:
            # Skip the item's file while searching
            if path == item_path:
                continue
            # Check for a matching filename
            if filename == ref:
                return relpath, None
            # Skip extensions that should not be considered text
            if os.path.splitext(filename)[-1] in settings.SKIP_EXTS:
                continue
            # Search for the reference in the file
            try:
                lines = linecache.getlines(path)
            except (SyntaxError, UnicodeDecodeError):
                log.trace("unable to read lines from: {}".format(path))  # type: ignore
                continue
            for lineno, line in enumerate(lines, start=1):
                if regex.search(line):
                    log.debug("found ref: {}".format(relpath))
                    return relpath, lineno

        msg = "external reference not found: {}".format(ref)
        raise DoorstopError(msg)

    @staticmethod
    def find_file_reference(ref_path, root, tree, item_path, keyword=None):
        """Find the external file reference.

        :raises: :class:`~doorstop.common.DoorstopError` when no
            reference is found

        :return: Tuple (ref_path, line) when reference is found

        """

        log.debug("searching for ref '{}'...".format(ref_path))
        ref_full_path = os.path.normpath(os.path.join(root, ref_path))

        for path, _filename, relpath in tree.vcs.paths:
            # Skip the item's file while searching
            if path == item_path:
                continue
            if path == ref_full_path:
                if keyword is None:
                    return relpath, None

                # Search for the reference in the file
                try:
                    lines = linecache.getlines(path)
                except SyntaxError:
                    log.trace("unable to read lines from: {}".format(path))  # type: ignore
                    continue

                log.debug("searching for ref '{}'...".format(keyword))
                pattern = r"(\b|\W){}(\b|\W)".format(re.escape(keyword))
                log.trace("regex: {}".format(pattern))  # type: ignore
                regex = re.compile(pattern)
                for lineno, line in enumerate(lines, start=1):
                    if regex.search(line):
                        log.debug("found ref: {}".format(relpath))
                        return relpath, lineno

        msg = "external reference not found: {}".format(ref_path)
        raise DoorstopError(msg)



# SPDX-License-Identifier: LGPL-3.0-only

"""Plug-in module to store requirements in a Mercurial repository."""

from doorstop import common
from doorstop.core.vcs.base import BaseWorkingCopy

log = common.logger(__name__)


class WorkingCopy(BaseWorkingCopy):
    """Mercurial working copy."""

    DIRECTORY = ".hg"
    IGNORES = (".hgignore",)

    def lock(self, path):
        log.debug("`hg` does not support locking: {}".format(path))
        self.call("hg", "pull", "-u")

    def edit(self, path):
        self.call("hg", "add", path)

    def add(self, path):
        self.call("hg", "add", path)

    def delete(self, path):
        self.call("hg", "remove", path, "--force")

    def commit(self, message=None):
        message = message or input("Commit message: ")
        self.call("hg", "commit", "--message", message)
        self.call("hg", "push")



# SPDX-License-Identifier: LGPL-3.0-only

"""Base classes and decorators for the doorstop.core package."""

import abc
import functools
import os
from typing import Dict

import yaml

from doorstop import common, settings
from doorstop.common import DoorstopError, DoorstopInfo, DoorstopWarning

log = common.logger(__name__)


def add_item(func):
    """Add and cache the returned item."""

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        item = func(self, *args, **kwargs) or self
        if settings.ADDREMOVE_FILES and item.tree:
            item.tree.vcs.add(item.path)
        # pylint: disable=W0212
        if item not in item.document._items:
            item.document._items.append(item)
        if settings.CACHE_ITEMS and item.tree:
            item.tree._item_cache[item.uid] = item
            log.trace("cached item: {}".format(item))  # type: ignore
        return item

    return wrapped


def edit_item(func):
    """Mark the returned item as modified."""

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        item = func(self, *args, **kwargs) or self
        if settings.ADDREMOVE_FILES and item.tree:
            item.tree.vcs.edit(item.path)
        return item

    return wrapped


def delete_item(func):
    """Remove and expunge the returned item."""

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        item = func(self, *args, **kwargs) or self
        if settings.ADDREMOVE_FILES and item.tree:
            item.tree.vcs.delete(item.path)
        # pylint: disable=W0212
        if item in item.document._items:
            item.document._items.remove(item)
        if settings.CACHE_ITEMS and item.tree:
            item.tree._item_cache[item.uid] = None
            log.trace("expunged item: {}".format(item))  # type: ignore
        BaseFileObject.delete(item, item.path)
        return item

    return wrapped


def add_document(func):
    """Add and cache the returned document."""

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        document = func(self, *args, **kwargs) or self
        if settings.ADDREMOVE_FILES and document.tree:
            document.tree.vcs.add(document.config)
        # pylint: disable=W0212
        if settings.CACHE_DOCUMENTS and document.tree:
            document.tree._document_cache[document.prefix] = document
            log.trace("cached document: {}".format(document))  # type: ignore
        return document

    return wrapped


def edit_document(func):
    """Mark the returned document as modified."""

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        document = func(self, *args, **kwargs) or self
        if settings.ADDREMOVE_FILES and document.tree:
            document.tree.vcs.edit(document.config)
        return document

    return wrapped


def delete_document(func):
    """Remove and expunge the returned document."""

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        document = func(self, *args, **kwargs) or self
        if settings.ADDREMOVE_FILES and document.tree:
            document.tree.vcs.delete(document.config)
        # pylint: disable=W0212
        if settings.CACHE_DOCUMENTS and document.tree:
            document.tree._document_cache[document.prefix] = None
            log.trace("expunged document: {}".format(document))  # type: ignore
        try:
            os.rmdir(document.path)
        except OSError:
            # Directory wasn't empty
            pass
        return document

    return wrapped


class BaseValidatable(metaclass=abc.ABCMeta):
    """Abstract Base Class for objects that can be validated."""

    def validate(self, skip=None, document_hook=None, item_hook=None):
        """Check the object for validity.

        :param skip: list of document prefixes to skip
        :param document_hook: function to call for custom document
            validation
        :param item_hook: function to call for custom item validation

        :return: indication that the object is valid

        """
        valid = True
        # Display all issues
        for issue in self.get_issues(
            skip=skip, document_hook=document_hook, item_hook=item_hook
        ):
            if isinstance(issue, DoorstopInfo) and not settings.WARN_ALL:
                log.info(issue)
            elif isinstance(issue, DoorstopWarning) and not settings.ERROR_ALL:
                log.warning(issue)
            else:
                assert isinstance(issue, DoorstopError)
                log.error(issue)
                valid = False
        # Return the result
        return valid

    @abc.abstractmethod
    def get_issues(self, skip=None, document_hook=None, item_hook=None):
        """Yield all the objects's issues.

        :param skip: list of document prefixes to skip
        :param document_hook: function to call for custom document
            validation
        :param item_hook: function to call for custom item validation

        :return: generator of :class:`~doorstop.common.DoorstopError`,
                              :class:`~doorstop.common.DoorstopWarning`,
                              :class:`~doorstop.common.DoorstopInfo`

        """

    @property
    def issues(self):
        """Get a list of the item's issues."""
        return list(self.get_issues())


def auto_load(func):
    """Call self.load() before execution."""

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        self.load()
        return func(self, *args, **kwargs)

    return wrapped


def auto_save(func):
    """Call self.save() after execution."""

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        if self.auto:
            self.save()
        return result

    return wrapped


class BaseFileObject(metaclass=abc.ABCMeta):
    """Abstract Base Class for objects whose attributes save to a file.

    For properties that are saved to a file, decorate their getters
    with :func:`auto_load` and their setters with :func:`auto_save`.

    """

    auto = True  # set to False to delay automatic save until explicit save

    def __init__(self):
        self.path = None
        self.root = None
        self._data: Dict[str, str] = {}
        self._exists = True
        self._loaded = False

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.path == other.path

    def __ne__(self, other):
        return not self == other

    @staticmethod
    def _create(path, name):
        """Create a new file for the object.

        :param path: path to new file
        :param name: humanized name for this file

        :raises: :class:`~doorstop.common.DoorstopError` if the file
            already exists

        """
        if os.path.exists(path):
            raise DoorstopError("{} already exists: {}".format(name, path))
        common.create_dirname(path)
        common.touch(path)

    @abc.abstractmethod
    def load(self, reload=False):
        """Load the object's properties from its file."""
        # 1. Start implementations of this method with:
        if self._loaded and not reload:
            return
        # 2. Call self._read() and update properties here
        # 3. End implementations of this method with:
        self._loaded = True

    def _read(self, path):
        """Read text from the object's file.

        :param path: path to a text file

        :return: contexts of text file

        """
        if not self._exists:
            msg = "cannot read from deleted: {}".format(self.path)
            raise DoorstopError(msg)
        return common.read_text(path)

    @staticmethod
    def _load(text, path, **kwargs):
        """Load YAML data from text.

        :param text: text read from a file
        :param path: path to the file (for displaying errors)

        :return: dictionary of YAML data

        """
        return common.load_yaml(text, path, **kwargs)

    @abc.abstractmethod
    def save(self):
        """Format and save the object's properties to its file."""
        # 1. Call self._write() with the current properties here
        # 2. End implementations of this method with:
        self._loaded = False
        self.auto = True

    def _write(self, text, path):
        """Write text to the object's file.

        :param text: text to write to a file
        :param path: path to the file

        """
        if not self._exists:
            raise DoorstopError("cannot save to deleted: {}".format(self))
        common.write_text(text, path, end=settings.WRITE_LINESEPERATOR)

    @staticmethod
    def _dump(data):
        """Dump YAML data to text.

        :param data: dictionary of YAML data

        :return: text to write to a file

        """
        return yaml.dump(data, default_flow_style=False, allow_unicode=True)

    # properties #############################################################

    @property
    def relpath(self):
        """Get the item's relative path string."""
        assert self.path
        relpath = os.path.relpath(self.path, self.root)
        return "@{}{}".format(os.sep, relpath)

    # extended attributes ####################################################

    @property  # type: ignore
    @auto_load
    def extended(self):
        """Get a list of all extended attribute names."""
        names = []
        for name in self._data:
            if not hasattr(self, name):
                names.append(name)
        return sorted(names)

    @auto_load
    def get(self, name, default=None):
        """Get an extended attribute.

        :param name: name of extended attribute
        :param default: value to return for missing attributes

        :return: value of extended attribute

        """
        if hasattr(self, name):
            cname = self.__class__.__name__
            msg = "'{n}' can be accessed from {c}.{n}".format(n=name, c=cname)
            log.trace(msg)  # type: ignore
            return getattr(self, name)
        else:
            return self._data.get(name, default)

    @auto_load
    @auto_save
    def set(self, name, value):
        """Set an extended attribute.

        :param name: name of extended attribute
        :param value: value to set

        """
        if hasattr(self, name):
            cname = self.__class__.__name__
            msg = "'{n}' can be set from {c}.{n}".format(n=name, c=cname)
            log.trace(msg)  # type: ignore
            setattr(self, name, value)
        else:
            self._data[name] = value

    # actions ################################################################

    def delete(self, path):
        """Delete the object's file from the file system."""
        if self._exists:
            log.info("deleting {}...".format(path))
            common.delete(path)
            self._loaded = False  # force the object to reload
            self._exists = False  # but, prevent future access
        else:
            log.warning("already deleted: {}".format(self))



# SPDX-License-Identifier: LGPL-3.0-only

"""Shared functions for the `doorstop.cli` package."""

import ast
import logging
import os
from argparse import ArgumentTypeError

from doorstop import common, settings

log = common.logger(__name__)


class capture:  # pylint: disable=R0903
    """Context manager to catch :class:`~doorstop.common.DoorstopError`."""

    def __init__(self, catch=True):
        self.catch = catch
        self._success = True

    def __bool__(self):
        return self._success

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type and issubclass(exc_type, common.DoorstopError):
            self._success = False
            if self.catch:
                log.error(exc_value)
                return True
        return False


def configure_logging(verbosity=0):
    """Configure logging using the provided verbosity level (0+)."""
    assert common.PRINT_VERBOSITY == 0
    assert common.STR_VERBOSITY == 3
    assert common.MAX_VERBOSITY == 4

    # Configure the logging level and format
    if verbosity == -1:
        level = settings.QUIET_LOGGING_LEVEL
        default_format = settings.DEFAULT_LOGGING_FORMAT
        verbose_format = settings.LEVELED_LOGGING_FORMAT
    elif verbosity == 0:
        level = settings.DEFAULT_LOGGING_LEVEL
        default_format = settings.DEFAULT_LOGGING_FORMAT
        verbose_format = settings.LEVELED_LOGGING_FORMAT
    elif verbosity == 1:
        level = settings.VERBOSE_LOGGING_LEVEL
        default_format = settings.DEFAULT_LOGGING_FORMAT
        verbose_format = settings.LEVELED_LOGGING_FORMAT
    elif verbosity == 2:
        level = settings.VERBOSE2_LOGGING_LEVEL
        default_format = verbose_format = settings.VERBOSE_LOGGING_FORMAT
    elif verbosity == 3:
        level = settings.VERBOSE3_LOGGING_LEVEL
        default_format = verbose_format = settings.VERBOSE_LOGGING_FORMAT
    else:
        level = settings.VERBOSE3_LOGGING_LEVEL
        default_format = verbose_format = settings.VERBOSE2_LOGGING_FORMAT

    # Set a custom formatter
    if not logging.root.handlers:
        logging.basicConfig(level=level)
        logging.captureWarnings(True)
        formatter = common.WarningFormatter(default_format, verbose_format)
        logging.root.handlers[0].setFormatter(formatter)

    # Warn about excessive verbosity
    if verbosity > common.MAX_VERBOSITY:
        msg = "maximum verbosity level is {}".format(common.MAX_VERBOSITY)
        logging.warning(msg)
        common.verbosity = common.MAX_VERBOSITY
    else:
        common.verbosity = verbosity


def configure_settings(args):
    """Update settings based on the command-line options."""

    # Parse common settings
    if args.no_reformat is not None:
        settings.REFORMAT = args.no_reformat is False
    if args.reorder is not None:
        settings.REORDER = args.reorder is True
    if args.no_level_check is not None:
        settings.CHECK_LEVELS = args.no_level_check is False
    if args.no_ref_check is not None:
        settings.CHECK_REF = args.no_ref_check is False
    if args.no_child_check is not None:
        settings.CHECK_CHILD_LINKS = args.no_child_check is False
    if args.strict_child_check is not None:
        settings.CHECK_CHILD_LINKS_STRICT = args.strict_child_check is True
    if args.no_suspect_check is not None:
        settings.CHECK_SUSPECT_LINKS = args.no_suspect_check is False
    if args.no_review_check is not None:
        settings.CHECK_REVIEW_STATUS = args.no_review_check is False
    if args.no_cache is not None:
        settings.CACHE_DOCUMENTS = args.no_cache is False
        settings.CACHE_ITEMS = args.no_cache is False
        settings.CACHE_PATHS = args.no_cache is False
    if args.warn_all is not None:
        settings.WARN_ALL = args.warn_all is True
    if args.error_all is not None:
        settings.ERROR_ALL = args.error_all is True

    # Parse `add` settings
    if hasattr(args, "server") and args.server is not None:
        settings.SERVER_HOST = args.server
    if hasattr(args, "port") and args.port is not None:
        settings.SERVER_PORT = args.port

    # Parse `publish` settings
    if hasattr(args, "no_child_links") and args.no_child_links is not None:
        settings.PUBLISH_CHILD_LINKS = args.no_child_links is False
    if hasattr(args, "no_levels") and args.no_levels is not None:
        settings.PUBLISH_BODY_LEVELS = False
        settings.PUBLISH_HEADING_LEVELS = args.no_levels != "all"


def literal_eval(literal, error=None, default=None):
    """Convert an literal to its value.

    :param literal: string to evaulate
    :param error: function to call for errors
    :param default: default value for empty inputs
    :return: Python literal

    >>> literal_eval("False")
    False

    >>> literal_eval("[1, 2, 3]")
    [1, 2, 3]

    """
    try:
        return ast.literal_eval(literal) if literal else default
    except (SyntaxError, ValueError):
        msg = "invalid Python literal: {}".format(literal)
        if error:
            error(msg)
        else:
            log.critical(msg)
        return None


def get_ext(args, error, ext_stdout, ext_file, whole_tree=False):
    """Determine the output file extensions from input arguments.

    :param args: Namespace of CLI arguments
    :param error: function to call for CLI errors
    :param ext_stdout: default extension for standard output
    :param ext_file: default extension for file output
    :param whole_tree: indicates the path is a directory for the whole tree

    :return: chosen extension

    """
    path = args.path if hasattr(args, "path") else None
    ext = None

    # Get the default argument from a provided output path
    if path:
        if whole_tree:
            ext = ext_file
        else:
            if os.path.isdir(path):
                error("given a prefix, [path] must be a file, not a directory")
            ext = os.path.splitext(path)[-1]
        log.debug("extension based on path: {}".format(ext or None))

    # Override the extension if a format is specified
    for _ext, option in {
        ".txt": "text",
        ".md": "markdown",
        ".html": "html",
        ".yml": "yaml",
        ".csv": "csv",
        ".xlsx": "xlsx",
        ".tex": "latex",
    }.items():
        try:
            if getattr(args, option):
                ext = _ext
                log.debug("extension based on override: {}".format(ext))
                break
        except AttributeError:
            continue
    else:
        if not ext:
            if path:
                error("given a prefix, [path] must include an extension")
            else:
                ext = ext_stdout
            log.debug("extension based on default: {}".format(ext))

    return ext


def show(message, flush=False):
    """Print (optionally flushed) text to the display.

    :param message: text to print
    :param flush: indicates the message is progress text

    """
    # show messages when enabled
    if common.verbosity >= common.PRINT_VERBOSITY:
        # unless they are progress messages and logging is enabled
        if common.verbosity == 0 or not flush:
            print(message, flush=flush)


def ask(question, default=None):
    """Display a console yes/no prompt.

    :param question: text of yes/no question ending in '?'
    :param default: 'yes', 'no', or None (for no default)

    :return: True = 'yes', False = 'no'

    """
    valid = {"yes": True, "y": True, "no": False, "n": False}
    prompts = {"yes": " [Y/n] ", "no": " [y/N] ", None: " [y/n] "}

    prompt = prompts.get(default, prompts[None])
    message = question + prompt

    while True:
        try:
            choice = input(message).lower().strip() or default
        except KeyboardInterrupt as exc:
            print()
            raise exc from None  # pylint: disable=raising-bad-type
        try:
            return valid[choice]
        except KeyError:
            options = ", ".join(sorted(valid.keys()))
            print("valid responses: {}".format(options))


def positive_int(value):
    """Evaluate a value as positive.

    :param value: passed in value to Evaluate

    :return: value casted to an integer

    """
    exc = ArgumentTypeError("'{}' is not a positive int value".format(value))
    try:
        ival = int(value)
    except ValueError:
        raise exc from None  # pylint: disable=raising-bad-type
    else:
        if ival < 1:
            raise exc
        return ival
